
#include	<i86.h>
#include	<string.h>
#include	"vesa.h"

/****************************************************************************/
/* VESA.C	-	Simple VESA driver.											*/
/* Written by Phil Jones <fil@muon.demon.co.uk>.							*/
/* Freely redistributable and all that.										*/
/* This code is a cut down and simplified version of my VESA driver.		*/
/* There is quite a bit of room for improvements and optimizations.			*/
/* Thanks should, however, go to Timo Talasmaa <timo@mits.mdata.fi> for		*/
/* pointing out a bug in my original screen copy routine.					*/
/****************************************************************************/

/****************************************************************************/
/* Private function prototypes.												*/
/****************************************************************************/

static void far * allocate_dos_memory(RealPointer * rp,ULONG bytes_to_allocate);
static void free_dos_memory(RealPointer * rp);

/****************************************************************************/
/* Private data.															*/
/****************************************************************************/

RealPointer				vesa_info_rp;
RealPointer				vesa_mode_data_rp;

/****************************************************************************/
/* Global data.																*/
/****************************************************************************/

VesaInfo		far	*	vesa_info			= NULL;
VesaModeData	far	*	vesa_mode_data		= NULL;
SLONG					vesa_granularity;
SLONG					vesa_page;
SLONG					vesa_width;
SLONG					vesa_height;
SLONG					vesa_bits_per_pixel;

/****************************************************************************/
/* Get information about the VESA driver.									*/
/*																			*/
/* Returns:																	*/
/*	TRUE		-	VESA driver present.  vesa_info set to point to a		*/
/*					a structure containing capability info etc.				*/
/*	FALSE		-	No VESA driver.											*/
/*																			*/
/****************************************************************************/

BBOOL vesa_get_info(void)
{
	union		REGS	regs;
	struct		SREGS	sregs;
	RMREGS				rmregs;
	BBOOL				ok			= FALSE;

	if((vesa_info = (VesaInfo far *)allocate_dos_memory(&vesa_info_rp,sizeof(VesaInfo))) != NULL)
	{
		memset(&rmregs,0,sizeof(rmregs));
		memset(&regs,0,sizeof(regs));
		memset(&sregs,0,sizeof(sregs));
		segread(&sregs);
		rmregs.eax = 0x4f00;
		rmregs.es = vesa_info_rp.Segment;
		rmregs.ds = vesa_info_rp.Segment;
		regs.x.eax = 0x300;
		regs.x.ebx = 0x10;
		regs.x.ecx = 0;
		sregs.es = FP_SEG(&rmregs);
		regs.x.edi = FP_OFF(&rmregs);
		int386x(0x31,&regs,&regs,&sregs);	//Get vesa info.
		if(rmregs.eax == 0x4f)
		{
			ok = TRUE;
		}
	}
	return(ok);
}

/****************************************************************************/
/* Free the information allocated with vesa_get_info()						*/
/****************************************************************************/

void vesa_free_info(void)
{
	free_dos_memory(&vesa_info_rp);
}

/****************************************************************************/
/* Set VESA video mode.  You MUST have previously called vesa_get_info().	*/
/*																			*/
/* Inputs:																	*/
/*	vmode	-	The VESA mode that you want. e.g. 0x101 is 640x480 256		*/
/*				colours.													*/
/*																			*/
/* Returns:																	*/
/*	TRUE	-	The mode has been set.										*/
/*	FALSE	-	The mode has not been set, probably not supported.			*/
/*																			*/
/****************************************************************************/

BBOOL vesa_set_mode(UWORD vmode)
{
	union	REGS	regs;
	BBOOL			ok = FALSE;

	memset(&regs,0,sizeof(regs));
	regs.w.ax = 0x4f02;
	regs.w.bx = vmode;
	int386(0x10,&regs,&regs);
	if(regs.w.ax == 0x004f)
	{
		vesa_get_mode_info(vmode);
		vesa_set_page(0);			//Probably not needed, but here for completeness.
		ok = TRUE;
	}
	return(ok);
}

/****************************************************************************/
/* Get VESA mode information.  Information is loaded into vesa_mode_data.	*/
/* vesa_get_info() must be called before calling this function.				*/
/* Should error check this really.											*/
/*																			*/
/* Inputs:																	*/
/*	vmode	-	The mode that you wish to get information about.			*/
/*																			*/
/* Returns:																	*/
/*	TRUE	-	vesa_mode_data points to a filled VesaModeData structure.	*/
/*	FALSE	-	vesa_mode_data probably invalid.  Mode probably not			*/
/*				supported.													*/
/*																			*/
/****************************************************************************/

BBOOL vesa_get_mode_info(UWORD vmode)
{
	union	REGS	regs;
	struct	SREGS	sregs;
	RMREGS			rmregs;
	BBOOL			ok		= FALSE;

	if((vesa_mode_data = (VesaModeData far *)allocate_dos_memory(&vesa_mode_data_rp,sizeof(VesaModeData))) != NULL)
	{
		memset(&rmregs,0,sizeof(rmregs));
		rmregs.es = vesa_mode_data_rp.Segment;
		rmregs.ds = vesa_mode_data_rp.Segment;
		rmregs.edi = 0;
		rmregs.eax = 0x4f01;
		rmregs.ecx = vmode;
		memset(&regs,0,sizeof(regs));
		memset(&sregs,0,sizeof(sregs));
		segread(&sregs);
		regs.x.eax = 0x300;
		regs.x.ebx = 0x10;
		regs.x.edi = (ULONG)&rmregs;
		int386x(0x31,&regs,&regs,&sregs);
		if(regs.h.al == 0)
		{

			//cache a few important items in protected mode memory area.

			vesa_granularity	= vesa_mode_data->WindowGranularity;
			vesa_width			= vesa_mode_data->PixelWidth;
			vesa_height			= vesa_mode_data->PixelHeight;
			vesa_bits_per_pixel	= vesa_mode_data->BitsPerPixel;
			ok = TRUE;
		}
	}
	return(ok);
}

/****************************************************************************/
/* Free the info previously allocated with vesa_get_info()					*/
/****************************************************************************/

void vesa_free_mode_info(void)
{
	free_dos_memory(&vesa_mode_data_rp);
}

/****************************************************************************/
/* Set the current vesa screen page.										*/
/*																			*/
/* Inputs:																	*/
/*	vpage	-	Page number to set.											*/
/*																			*/
/****************************************************************************/

void vesa_set_page(SLONG vpage)
{
	union	REGS	regs;

	vesa_page = vpage;
	regs.w.ax = 0x4f05;
	regs.w.bx = 0;
	regs.w.cx = 0;
	regs.w.dx = (UWORD)(vpage == 0 ? 0 : (UWORD)((vpage * 64) / vesa_granularity));
	int386(0x10,&regs,&regs);
	regs.w.ax = 0x4f05;
	regs.w.bx = 1;
	regs.w.cx = 0;
	regs.w.dx = (UWORD)(vpage == 0 ? 0 : (UWORD)((vpage * 64) / vesa_granularity));
	int386(0x10,&regs,&regs);
}

/****************************************************************************/
/* Generic screen copy.														*/
/*																			*/
/* Inputs:																	*/
/*	source		-	Your source screen buffer.								*/
/*	swidth		-	The width of the source/destination screen.				*/
/*	sheight		-	The height of the source/destination screen.			*/
/*	bpp			-	The number of bytes per pixel of the source/destination	*/
/*					screen.													*/
/*																			*/
/****************************************************************************/

void vesa_screen_swap(UBYTE * source,SLONG swidth,SLONG sheight,SLONG bpp)
{
	void	*	dest;
	SLONG		page		= 0;
	SLONG		scr_size	= (swidth * sheight * bpp);	//width * height * bytes per pixel.

	dest = (void *)source;
	while(scr_size > 0)
	{
		vesa_set_page(page);
		memcpy((void *)0xa0000,dest,(scr_size > 65536 ? 65536 : scr_size));
		scr_size -= 65536;
		dest = (void *)((ULONG)dest + 65536);
		page++;
	}

}

/****************************************************************************/
/* Allocate a region of DOS memory.											*/
/****************************************************************************/

static void far * allocate_dos_memory(RealPointer * rp,ULONG bytes_to_allocate)
{
	void	far	*		ptr		= NULL;
	union	REGS		regs;

	bytes_to_allocate = ((bytes_to_allocate + 15) & 0xfffffff0);	//Round up to nearest paragraph.
	memset(&regs,0,sizeof(regs));
	regs.w.ax = 0x100;
	regs.w.bx = (UWORD)(bytes_to_allocate >> 4);		//Allocate dos memory in pages, so convert bytes to pages.
	int386(0x31,&regs,&regs);
	if(regs.x.cflag == 0)
	{	//Everything OK.
		rp->Segment = regs.w.ax;
		rp->Selector = regs.w.dx;
		ptr = MK_FP(regs.w.dx,0);
	}
	return(ptr);
}

/****************************************************************************/
/* Free an area of DOS memory.												*/
/****************************************************************************/

static void free_dos_memory(RealPointer * rp)
{
	union	REGS		regs;

	regs.w.ax = 0x101;
	regs.w.dx = rp->Selector;
	int386(0x31,&regs,&regs);
}

/****************************************************************************/
/* END																		*/
/****************************************************************************/

